/*
 * Copyright (C) The Apache Software Foundation. All rights reserved.
 *
 * This software is published under the terms of the Apache Software
 * License version 1.1, a copy of which has been included with this
 * distribution in the ASF-LICENSE.txt file.
 */

package com.gregor.util;

import java.beans.BeanInfo;
import java.beans.Introspector;
import java.beans.IntrospectionException;
import java.beans.PropertyDescriptor;

import java.lang.reflect.Method;


/**
   Used for inferring configuration information for a log4j's component.

   @author  Anders Kristensen
 */
public class PropertyGetter {
  protected static final Object[] NULL_ARG = new Object[] {};
  protected Object obj;
  protected PropertyDescriptor[] props;

  public interface PropertyCallback {
    void foundProperty(Object obj, String prefix, String name, Object value);
  }

    public class PropertyByName implements PropertyCallback {
	public boolean found = false;
	public Object value = null;
	public String name = null;

	public void foundProperty(Object obj, String prefix, String name,
			   Object value) {

	    if (this.name.equals(name)) {
		this.value = value;
		this.found = true;
	    }
	}
	
    }

  /**
    Create a new PropertyGetter for the specified Object. This is done
    in prepartion for invoking {@link
    #getProperties(PropertyGetter.PropertyCallback, String)} one or
    more times.

    @param obj the object for which to set properties */
  public
  PropertyGetter(Object obj) throws IntrospectionException {
    BeanInfo bi = Introspector.getBeanInfo(obj.getClass());
    props = bi.getPropertyDescriptors();
    this.obj = obj;
  }

  public
  static
  void getProperties(Object obj, PropertyCallback callback, String prefix) {
    try {
      new PropertyGetter(obj).getProperties(callback, prefix);
    } catch (IntrospectionException ex) {
      // ignore
    }
  }

  public
  void getProperties(PropertyCallback callback, String prefix) {
    for (int i = 0; i < props.length; i++) {
      Method getter = props[i].getReadMethod();
      if (getter == null) continue;
      if (!isHandledType(getter.getReturnType())) {
	//System.err.println("Ignoring " + props[i].getName() +" " + getter.getReturnType());
	continue;
      }
      String name = props[i].getName();
      try {
	Object result = getter.invoke(obj, NULL_ARG);
	//System.err.println("PROP " + name +": " + result);
	if (result != null) {
	  callback.foundProperty(obj, prefix, name, result);
	}
      } catch (Exception ex) {
	// ignore
      }
    }
  }

    /**
     * Gets a property value from an object.
     *
     * Calls a get* method to retrieve a property value from an object.
     * Returns null if the getter method does not exist for the requested
     * property value.
     *
     * @param obj Object
     * @param name Property value to retrieve
     * @return property value
     * @throws IntrospectionException If an error occurs during introspection
     */
    public static String getProperty(Object obj, String name)
           throws IntrospectionException {
	return new PropertyGetter(obj).getProperty(name);
    }

    /**
     * Gets a property value.
     *
     * Calls a get* method to retrieve a property value from an object.
     * Returns null if the getter method does not exist for the requested
     * property value.
     *
     * @param name Property value to retrieve
     * @return property value
     * @throws IntrospectionException If an error occurs during introspection
     */
    public String getProperty(String name) throws IntrospectionException {
	PropertyByName callback = new PropertyByName();

	callback.name = name;

	getProperties(callback, "");
	
	return callback.found ? callback.value.toString() : null;
    }


  protected
  boolean isHandledType(Class type) {
    return String.class.isAssignableFrom(type) ||
      Integer.TYPE.isAssignableFrom(type) ||
      Long.TYPE.isAssignableFrom(type)    ||
      Boolean.TYPE.isAssignableFrom(type);
  }
}
